
/* Tenison Technology ARM Assembler */

/*
 * This file: carmas.c                 TTRX ARMBOS MAR96
 * 
 *  Jul 98 defs limit fixed. single quotes. mrs cpsr fixed.
 *         copy augmented
 * 
 * Software from Tenison Technology
 * (C) 1990-1998 Tenison Technology 
 * info@tenisontech.com.  Fax +44 1223 565917
 *
 * No responsibility can be accepted for loss or consequential loss.
 * Not for commercial use without license.  
 *
 * 
 */


 
char *assembler = "CARM";
 
#include <stdio.h>
#include "carmas.h"
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>
 
extern void lok();       // Two entry points to 68k specific code.
extern void initspecific();

#define  OBJMAX  1500 /* IHEX max line */
#define  labmax  25  /* Max label length */
#define  linemax 400 /* longest input line */


void objword();
void objbyte();
int eval3();
int eval2();
int ledf = 0;    /* link editable output file flag */
int verbosef = 0;
int listfirst = 0;
int filterbssf = 0;
int binaryf = 0;
int conditional_eval_code;
char lable[OBJMAX], inst[OBJMAX], token[OBJMAX], etoken[OBJMAX];
int cards;        /* Total lines read */
int line  ;      /* line number */
char *currentsrc, *oldsrc;

static int copyf; /* Zero during null'd parts of conditional assembly */
int baddr[2]  ;      /* object address vector */
int seclen[2];       /* length of text and bss sections */
int object[OBJMAX];      /* bcpl string containing object bytes */
int pass;        /* 1 or 2 */

char *jobname;   /* string containing filename */
int eos, running;/* boolean flags end of src */
int errors    ;  /* error counter */
char *error_type;  /* zero for Ok */
int tab;  /* index for inst array */
int tvalue;  /* type of instruction */
char ech;  /* as ch but for eval */
int imed;  /* boolean set by # sign */
int oldline;     /* line in primary src file */
char * gnem; /* Current nemonic info line */
int readonlyopt = 0; /* Set if object file must not update itself */

struct stnode *wordnode, **spoi, **stab_start, *lhsnode;

void writeline(FILE *ss);


FILE  *srccis, *oldcis = NULL, *objcis, *listcis;

union
{
  int i; char *s;
} error_sup;  /* Additional error information */

int xaddr     ;  /* memory allocation skip */
int error_tab ;  /* point of error */
int equate    ;  /* boolean flag for address print */

int relorigin ;  /* boolean flag for relocatable code */
int equateval ;  /* Value assigned in an equate */
int output = O_TEXT;  /* Type of code being generated */

void error(char *e, char *s)
{ 
  /*
  writeline(stdout);
  printf("e %s\n", e);
  */
  if (error_type == NULL)
    {
      error_type = e;
      error_tab = tab;
      error_sup.s = s;
    }
}

void error_s(char *e, char *s)
{ 
  if (error_type == NULL)
    {
      error_type = e;
      error_tab = tab;
      error_sup.s = s;
    }
}

char *inpoi;
char inbuffer[linemax];

char mrdch()
{
  char c = *(inpoi++);
  if (c == 9) c = ' ';
  return c;
}


void readline() 
{
  int i;
  char ch;
  lable[0] = (char) 0;
  inst[0] = (char) 0;
 
  if (feof(srccis))
    {
      inbuffer[0] = 0;
      eos = 1;
      return;
    }
  if (fgets(inbuffer, linemax, srccis))
    {
      inpoi = inbuffer;
      ch = mrdch();
      cards ++;
      line = line + 1;
      if (ch == '\n')  
	{
	  return;
	}
      
      /* read lable */
      if (ch !=' ' && ch != ';' && ch != '\n' && ch != '*')
	{  
	  i = 0;
	  while (isalnum(ch) || ch == '_')
	    {         
	      if (ch != (char) 13) lable[i++] = ch;
	      ch = mrdch(); 
	      if (i >= labmax)  
		{
		  error("label too long", 0);
		  break;
		}
	    }
	  lable[i++] = (char) 0;
	}
      if (ch == ':') ch = mrdch();
      
      while (ch==' ')  ch = mrdch();  /* read instruction or comment */
      
      i = 0;
      while (ch !='\n' && ch != EOF)
	{  
	  if (ch != (char) 13) inst[i++] = ch;
	  ch = mrdch();
	  if (i > OBJMAX)  break;
	}
      inst[i++] = (char) 0;
    }
}


void newline() 
{ 
  printf("\n");
}

void writeline(FILE *ss) 
{
  int i;
  fprintf(ss, "%4i ", line);
  
  if (lable[0] ==';')
    { 
      fprintf(ss, "%s %s\n", lable, inst);
      return;
    }
  
  if (object[0] > 0)  
    { 
      int j = (object[0] > 4) ? 4 : object[0];
      fprintf(ss, "%06X ", baddr[output]);
      
      for (i=1; i<=j; i++) fprintf(ss, "%02X", object[i]);
      for (i=j; i<=4; i++) fprintf(ss, "  ");
      fprintf(ss, " ");
    }
  else if (equate)  fprintf(ss, "=%08X         ", equateval);
  else fprintf(ss, "              ");
  
  fprintf(ss, "%25s", lable);
  fprintf(ss, "  %s\n", inst);
  
  if (object[0] > 4)
    {
      fprintf(ss, "            ");
      for (i=5; i <= object[0]; i++)
	{
	  fprintf(ss, "%02X", object[i]);
	}
      fprintf(ss, "\n");
    }
}


int hash(char *s)
{ 
  int r=0;
  while(*s)
  {
    r = r + r + *s;
    s++;
  }
  return r & (HASHSIZE-1);
}

int vlookup(char *lab, int insert)               /* search symbol table */
{  
  int h = hash(lab);
  wordnode = stab_start[h];
  while (wordnode)
    { 
      /* printf("<%s> <%s> %i %i \n", lab, wordnode->textual, hash(lab), hash(wordnode->textual));  */
      
      if (strcmp(lab, wordnode->textual)==0)  return 1;
      wordnode = wordnode->stnext;
    }
  if (insert)
    {
      wordnode = NEW(STNODE);
      wordnode->stnext = stab_start[h];
      stab_start[h] = wordnode;
      wordnode -> textual = strdup(lab);
    }
  return 0;
}


void say_error(FILE *ss)
{ 
  int i;
  errors = errors + 1;
  fprintf(ss, "ERROR------------------------------------------");
  for (i=1; i<= error_tab; i++)  putc('-', ss);
  fprintf(ss, "^\n'%s': ", currentsrc);
  fprintf(ss, error_type, error_sup);
  fprintf(ss, "\n");
}

char nextch(int skip) 
{
  ech = inst[tab++];
  if (skip && (ech == ' ' || ech == (char) 9))  nextch(1);       /* skip leading spaces (recursively) */
  return ech;
}

void ignblank()
{
  while (ech == ' ' || ech == (char) 9) nextch(1);
}

void chcomma() 
{
  char line[2];
  line[0] = ech;
  line[1] = 0;
  if (ech!=',')  error("Comma expected not `%s`", strdup(line));
  nextch(1);
}

void checkfor(char c)
{
  char p[2];
  if (ech != c)  
    {
      p[0] = c;
      p[1] = 0;
      error("<%s> expected.", strdup(p));
  }
  nextch(1);
}

int aschex(char cch)
{ 
  int ch = toupper(cch);
  return  (ch < '0') ? -1:
    (ch > 'F') ? -1:
    (ch <= '9') ? ch - '0':
    (ch >= 'A') ? ch - 'A' + 10: -1;
}



/* Recursive descent expression evaluation */
int eval9(dummy) 
{
  if (ech == '$')                 /* HEX CONSTANT */
    { 
      int s=0;
      int df = 0;
      nextch(1);
      while (aschex(ech) >= 0)
	{ 
	  df = 1;
	  s = (s<<4) + aschex(ech);
	  nextch(1);
	}
      if (df == 0) error("Bad $ hex constant", 0);
      return s;
    }
  
  if (isdigit(ech))                  /* decimal constant - unless suffixed */
    { 
      int s = 0; 
      int base = 10;
      int starttab = tab;
      while (aschex(ech) >= 0)
	{  
	  s = s*base + aschex(ech);
	  nextch(0);
	  if (ech == 'X' || ech == 'x') 
	    { 
	      base = 16; nextch(1); 
	    }
	  
	}
      if (ech == 'H')  /* H as suffix, means redo again as base 16 from start */
	{
	  s = 0;
	  tab = starttab-1;
	  nextch(1);
	  while (aschex(ech) >= 0)
	    {  
	      s = s*16 + aschex(ech);
	      nextch(1);
	    }
	  nextch(1); /* skip H */
	}
      ignblank();
      return s;
    }
  
  if (ech == '-')                 /* monadic negate */
    {
      nextch(1);
      return 0-eval9(tab);
    }
  
  if (ech == '~')                 /* one's complement */
    {
      nextch(1);
      return -1-eval9(tab);
    }

  if (ech == '.')   /*assembly point for ARX is denoted with dot */
    {
      nextch(1);
      return baddr[output];
    }

    if (ech == '(')                /* expression in parens */
      {
	int r;
        nextch(1);
        r = eval2();
        checkfor(')');
        return r;
      }

    if (ech == '\'')                /* ascii quote char */
      { 
	int r = nextch(0);
        if (nextch(1) != '\'')  error("Mismatched quotes.", 0);
        nextch(1);
        return r;
      }
    
    
    if (ech== '#' || ech ==  '=')       /* immediate addressing flag */
      {
	int js;
	nextch(1);
        js = eval9(tab);
        imed = 1;
        return js;
      }


    if ((isalpha(ech)) || (ech=='_') )       /* symbol look up */
      {
	int i = 0;
        while (isalnum(ech) || ech == '_')
	  {
	    token[i++] = ech;
	    nextch(0);
	  }
        token[i++] = (char) 0;
        ignblank();
        if (vlookup(token, 0))
	  { 
	    int j = wordnode->type;   /* symbol found */

	    if (j==RELOCATABLE || j==EXTERN)
	      {
		if (*spoi) error("Double relocation required", 0);
		*spoi = wordnode;
	      }
	    if (j == EXTERN) return 0;   /* Do not give BSS offsets - ld adds them */
	    return wordnode->value;      /* return value */
          }
        else
	  { 
	    strcpy(etoken, token);
	    error_s("Symbol <%s> not defined.", etoken);
	    return 0;
	  }
      }
    error("Unknown thing in expression %s", strdup(inst+tab-1));
    return 0;
}


int eval5() 
{ 
  return  eval9(tab);
}

int eval4() 
{
  int s = eval5();
  while (ech == '*')  
     {  
       nextch(1) ;  
       s = s * eval5();  
     }
  return s;
 }


/*
 * Add and subtract.
 * Handles subtraction of two relocatable values.
 */
int eval3() 
{
  int s = eval4();
  while (ech == '+' || ech == '-')  
  { 
    STNODE *spoicopy = *spoi;
    char k = ech;
    if (k == '-') *spoi = NULL;
    
    nextch(1); 
    if (k == '+') s = s + eval4();  
    else 
      {
	s = s - eval4();  
	if (*spoi && spoicopy == NULL) error("negative relocation", "");
	if (*spoi == NULL) *spoi = spoicopy;
      }
  }
  return s;
}

/*
 * Left and right shift.
 */
int eval2() 
{
  int s = eval3();
  while (ech == '<' || ech == '>')  
  { 
    char k = ech;
    int shiftf = 0;
    nextch(1); 
    if (ech == k)
      {
	nextch(1);
	shiftf = 1;
      }

    if (shiftf)
      {
	if (k == '>')  s = s >> eval3();  
	else s = s << eval3();
	return s;
      }
    /* rest unwritten */
  }
  return s;
}

/* Check data size */
int sexmeasure(data) 
{
  int k = (data  >= 0) ? data: -1- data;
  return (k<128) ? 0:              /* byte result */
    (k<32768) ? 1:           /* word */
    2;                       /* long word */
}


int eval(STNODE **lspoi)
{  
  int res;
  spoi = lspoi;
  *spoi = NULL;
  res = eval2();
  return res;
}

/* IF big endian */
void big_objword(unsigned int w)
{ 
  objbyte(w >> 24);
  objbyte(w >> 16);
  objbyte(w >> 8);
  objbyte(w);
}

/* IF big endian */
void objword(unsigned int w)
{ 
  objbyte(w);
  objbyte(w >> 8);
  objbyte(w >> 16);
  objbyte(w >> 24);
}

void objbyte(b)
{
 
  if (filterbssf && output == O_BSS) 
    {
      object[0] += 1;
    }
  else
    {
      int x = object[0];
      object[++x] = b & 0xFF;
      if (x < OBJMAX) object[0] = x;
      else error("too much object", "");
    }
}


/* Note that an external or relocatable symbol has been used */
void symbnote(struct stnode *wn, int xvalue) 
{ 
  struct locchain *r;
  if (pass!=2) return;
  r = (struct locchain *) malloc(sizeof(struct locchain));
  r->next = wn->schain;
  wn->schain = r;
  r->val = xvalue;
}


void org(int a)
{
  relorigin = 0;
  baddr[O_TEXT] = a;
}

void padorg(int a)
{
  int d = a - baddr[output];
  if (d < 0) error("pad already passed", "");
  while(d > 0)
    {
      objword(0xC0C0C0C0);
      d -= 4;
    }
}

/* put a lable into symbol table */
void lable_handle(char sf)  
{
  if (vlookup(lable, 1))
    { 
      if (pass==1 && sf != 'e') 
	{
	  if (wordnode->type == EXTERN)   /* extern actually defined here */
	    { 
	      if (output == O_TEXT)
		{ 
		  wordnode->value = baddr[O_TEXT];
		  wordnode->type = (relorigin) ? RELOCATABLE: ABSOLUTE;
		}
	      else /* Convert extern to local bss */
		{
		  wordnode->value = baddr[O_BSS];
		  wordnode->equatef = EXTERN;
		  wordnode->type = EXTERN;
		  
		  if (objcis && ledf) fprintf(objcis, "-B %s %08X\n",  lable, baddr[O_BSS]);
		}
	    }
	  else wordnode-> type  = DMTO;   /* found twice on first pass, dmto */
	}
      else if (pass == 2)
	{
	  if (wordnode->type==DMTO)  error("Label %s defined more than once", strdup(wordnode->textual));
	  if (sf != 'e' && wordnode->value != baddr[output] && wordnode->equatef == 0)
	    {
	      printf("Label value changed for %s, old=%X new=%X\n", lable, wordnode->value, baddr[output]);
	      printf("type = %i, pass =%i, equatef=%i\n", wordnode->type, pass,
		     wordnode->equatef);
	    }
	}
    }
  else /* Not found before */
    { 
      if (sf == 'e') 
	{
	  wordnode->value = 0;
	  wordnode->type = EXTERN;
	  wordnode->equatef = EXTERN;
	}    
      
      else if (output == O_TEXT || ledf == 0)
	{
	  /* initially enter as a straightforward code lable */
	  wordnode->value = baddr[output]; 
	  wordnode->equatef = 0;           /* So clear equate flag           */
	  wordnode->type = (relorigin) ? RELOCATABLE: ABSOLUTE;
	}
      else if (output == O_BSS)
	{
	  /* If linking, enter bss labels as externs in symbol table */
	  wordnode->value = baddr[O_BSS];
	  wordnode->equatef = EXTERN;
	  wordnode->type = EXTERN;
	  if (objcis && !binaryf) fprintf(objcis, "-B %s %08X\n",  lable, baddr[O_BSS]);
	}
      
      wordnode->schain = NULL;
      wordnode->textual = (char *) malloc(strlen(lable)+1);
      strcpy(wordnode->textual, lable);
    }
}

void objhex(unsigned char c)
{
  char c1 = c >> 4;
  if (c1 > 9) c1 += 7;
  putc(c1 + '0', objcis);
  c1 = c & 15;
  if (c1 > 9) c1 += 7;
  putc(c1 + '0', objcis);
}


void ihex_code_output() 
{
  static int segment = 0;
  int i = 1;
  int la = baddr[output];
  int amount = object[0];
  
  if (filterbssf && output == O_BSS) return;
  if (objcis == NULL) return;
 
#if 0
  if (relorigin && xaddr && output == O_TEXT)  /* Fill uninitialised storage */
    {
      for (i=0; i< xaddr; i++)  fprintf(objcis, "00");
    }
#endif
  if (amount == 0) return;
  


  while (amount) 
    {
      int sum;

      int j = (amount >= 196) ? 196: amount;
      if (j + (la & 0xFFFF) > 0xFFFF) j = 0x10000 -  (la & 0xFFFF);
      amount -= j;

      if ((la & 0x000F0000) != segment)
	{
	  segment = la & 0x000F0000;
	  sum = (segment >> 12) + 2 + 2;
 	  fprintf(objcis, ":02000002%04X%02X\n", segment >> 4, (256 - sum) & 0xFF);
	  segment = la & 0x000F0000;
	}
      
     

      fprintf(objcis, ":%02X%04X00", j, la & 0xFFFF);
      sum = j + la + (la >> 8);
      

      la += j;
      while(j-- > 0)
	{
	  objhex(object[i]);
	  sum = object[i] + sum;
	  i += 1;
	}
      fprintf(objcis, "%02X\n", (256 - sum) & 255);
    }
}



void binary_code_output() 
{
  int i;
  if (objcis == NULL) return;
 

  if (relorigin && xaddr && output == O_TEXT)  /* Fill uninitialised storage */
    {
      for (i=0; i< xaddr; i++)  putc(0, objcis);
    }

  if (object[0] == 0) return;
  
  if (output != O_TEXT) error("code in bss", 0);
  
  
  for (i=1; i<=object[0]; i++) 
    {
      putc(object[i], objcis);
    }
}


void code_output() 
{
  if (binaryf) binary_code_output();
  else ihex_code_output();
}


void unstackcis()
{
  if (oldcis == NULL) running = 0;
  else
    { 
      srccis = oldcis;
      oldcis = NULL;
      line = oldline;
      eos = 0;
      currentsrc = oldsrc;
      readline();
    }
}

void perform_pass()
{ 
  char sf; running = 1;
  output = O_TEXT;
  eos = 0;
  copyf = 1;  /* Default to ON */
  cards = 0;
  seclen[0] = 0; /* No output text yet */
  seclen[1] = 0; /* No output text yet */
  /*  printf("Source: '%s' pass %i.\n", jobname, pass);
   */
  currentsrc = jobname;

  srccis = fopen(jobname, "r");
  if (srccis == NULL) 
    {
      printf("Can't open src file '%s' \n", jobname);
      perror(jobname);
      exit(1);
    }
  line = 0;
  relorigin = 1;  /* default to relocatable module */
  baddr[O_TEXT] = 0;                /* default origin */
  baddr[O_BSS] = 0x10000;           /* default bss origin */
  errors = 0;


  while (running)
    {
      error_type = NULL;
      object[0] = 0; 
      readline();
      if (eos) 
	{
	  unstackcis();
	  continue;
	}
      equate = 0;
      xaddr = 0;
      if (lable[0] == ';') continue;
      sf = (char) 0;
      
      if (copyf && inst[0] == '.')
	{
	  if (strncmp(inst, ".EXTERN", 6)==0) sf = 'e';
	  else if (strncmp(inst, ".GLOBAL", 6)==0) sf = 'g';
	  else if (strncmp(inst, ".extern", 6)==0) sf = 'e';
	  else if (strncmp(inst, ".global", 6)==0) sf = 'g';
	}
      
      if (copyf && lable[0] != (char) 0) 
	{
	  lable_handle(sf);
	  lhsnode = wordnode;
	}
      else lhsnode = NULL;
      
      tab = 0;
      lok(); 
      if ((error_type !=NULL) && (pass==2) && listcis != stdout)
        {  
          writeline(stdout);
          say_error(stdout);
        }
      
      if (copyf && listfirst && pass == 1) writeline(stdout);
      
      if (listcis != NULL)
        {
	  writeline(listcis);
	  if (error_type != NULL) say_error(listcis);
        }
      if (copyf)
	{
	  /* Round up to boundary 
	   * if (xaddr & 1) xaddr++;  
	   * if (object[0] & 1) objbyte(0); 
	   */
	  error_type = 0;
	  if (pass==2)  code_output();
	  if (error_type) say_error(stdout);
	  baddr[output] = baddr[output] + object[0] + xaddr;
	  seclen[output] += object[0];
	  if (relorigin) seclen[output] +=  xaddr;
	}
    }
  fclose(srccis);
}


int chop(o, c)           /* check presence of case in opts */
{ 
  return 0;
}

void list_externs(int x, char *title) 
{
  int h;
  if (objcis == NULL) return;
  fprintf(objcis, title);
  for (h=0; h<HASHSIZE; h++)
    {
      struct stnode *p = stab_start[h];
      while (p)
	{ 
	  if (p->type == x)
	    { 
	      struct locchain *q = p->schain;
	      if (q)  
		{
		  fprintf(objcis, ">%s ", p->textual);
		  while (q)
		    {
		      fprintf(objcis, "%08X ", q->val);
		      q = q->next;
		    }
		  fprintf(objcis, "\n");
		}
	    }
	  p = p->stnext;
	}
    }
}


int main(int argc, char **argv)
{  
  int lenfirst;
  int h;
  /* printf("Slider 68 series assemblers.  cbg. \n");
   */
  stab_start = (struct stnode **) malloc(HASHSIZE * 4);
  for (h=0; h<HASHSIZE; h++) stab_start[h] = NULL; 
  initspecific();
  if (argc < 2) 
    {
      printf("Usenix is: %s [ -rom -filterbss ] srcfile objfile [ listfile [ symboldumpfile]]\n", argv[0]);
      exit(1);
    }
  while(1)
    {

      if (strcmp(argv[1], "-rom")==0)
	{
	  readonlyopt = 1;
	  argv++; argc--;
	  continue;
	}


      if (strcmp(argv[1], "-listfirst")==0)
	{
	  listfirst = 1;
	  argv++; argc--;
	  continue;
	}


      if (strcmp(argv[1], "-filterbss")==0)
	{
	  filterbssf = 1;
	  argv++; argc--;
	  continue;
	}


      if (strcmp(argv[1], "-verbose")==0)
	{
	  verbosef = 1;
	  argv++; argc--;
	  continue;
	}

      if (strcmp(argv[1], "-binary")==0)
	{
	  binaryf = 1;
	  argv++; argc--;
	  continue;
	}

      break;
    }

  jobname = strdup(argv[1]);    
  /**====================== */
  /* FIRST PASS */
  /**====================== */
  pass = 1;
  objcis = NULL;
  if (fopen(argv[1], "r")==NULL)
    {
      printf("as: cant open src file %s\n", argv[1]);
      perror(argv[1]);
      exit(1);
    }
  /* Module file */
  if (argc > 2)
    { 
      objcis = fopen(argv[2], "w");
      if (objcis == NULL)
	{
	  perror(argv[2]);
	  exit(1);
	}
      if (!binaryf) fprintf(objcis, "$$ TT/CBG %s assembler job %s\n", assembler, jobname);
    }
  listcis = NULL;
  
  
  perform_pass();
  lenfirst = baddr[O_TEXT];
  
  /**====================== */
  /* 2nd   PASS */
  /**====================== */
  if (verbosef) printf("Second pass\n");
  if (seclen[0] + seclen[1] > 0)
    {
      pass = 2;
      
      if ((objcis && !binaryf && ledf))
	{
	  if (seclen[O_BSS]) fprintf(objcis, "-BL %08X\n", seclen[O_BSS]);     
	  if (ledf) fprintf(objcis, "-T%08X\n", seclen[O_TEXT]);
	  
	}
      if (argc > 3)
        { 
	  if (strcmp(argv[3], "-") == 0) listcis = stdout;
	
           else listcis = fopen(argv[3], "w");
	  if (listcis == NULL)
	    {
	      perror(argv[3]);
	      exit(1);
	    }
	  fprintf(listcis, "** CBG %s assembler job %s\n", assembler, jobname);
	}
      perform_pass();
      if (!binaryf && ledf)
	{
	  list_externs(RELOCATABLE, "\n-R\n");
	  list_externs(EXTERN, "\n-E\n");
	}
      if (lenfirst != baddr[O_TEXT])
        {
	  printf("as: %s unequal passes %x  %x \n", argv[1], lenfirst, baddr[O_TEXT]);
	  errors = errors + 1;
        }
    } /* endif textlen > 0 */

  if (listcis != NULL)
    {
      fprintf(listcis, "\n\nERRORS = %i,   SRC CARDS READ = %i\n\n",
	      errors, cards);
      if (listcis != stdout) fclose(listcis);
    }

  /* dump symbols optionally
   */
  if (argc > 4)
    {
      FILE * sss;
      if (!strcmp(argv[4],"-")) sss = stdout;
      else sss = fopen(argv[4], "w");
      if (verbosef) printf("Dumping symbols\n");
      if (listcis == NULL)
	{
	  perror(argv[4]);
	  exit(1);
	}
      for (h=0; h<HASHSIZE; h++)
	{
	  struct stnode *p = stab_start[h];
	  while (p)
	    {
	      fprintf(sss, "%08X %i %s\n", p->value, p->type, p->textual);
	      p = p->stnext;
	    }
	}
      fclose(sss);
    }
  

  /* Close module */
  if (objcis != NULL)
    {
      if (ledf) fprintf(objcis, "-%c \n", (errors>0) ? 'X': '-');
      fclose(objcis);
      if (errors) remove(argv[2]);
    }

  /* if (errors == 0) printf("Successful completion.\n");
   */
  return errors;
}



void gas_get()
{
  if (oldcis)
    {
      printf("Nested textual includes or GETs not supported.\n");
      exit(1);
    }
  oldcis = srccis;
  oldline  = line;
  oldsrc = currentsrc;
  {
    char *n = (char *) malloc(80);
    int i = 0;
    ignblank();
    while (ech != (char) 0 && ech != ';')
      {
	if (ech != (char) 0x22) n[i++] = ech;
	nextch(1);
      }
    n[i++] = (char) 0;
    srccis = fopen(n, "r");
    if (srccis == NULL)
      { 
	printf("Cant open included file '%s'.\n", n);
	exit(1);
      }
    /*   if (pass==1) printf("Reading: %s\n", n);
     */
    currentsrc = n;
  }
} 

void gas_copy(int f)
{
  copyf = f;
}


IP *iphashtab[255];

typedef struct parseitem 
{
  int dpp, mod, val, val1, offset, sign, shift;
  struct stnode *symb;
} PI;





char *conditions[14] =
{
  "EQ",
  "NE",
  "CS",
  "CC",
  "MI",
  "PL",
  "VS",
  "VC",
  "HI",
  "LS",
  "GE",
  "LT",
  "GT",
  "LE" 
};

int iscondition(char *p)
{
  int i;
  for (i=0;i<14;i++)
    {
      if (p[0] == conditions[i][0] && p[1] == conditions[i][1]) 
	{
	  conditional_eval_code = (i << 28);
	  break;
	}
    }
  return i;  /* 14 on exit, which is the ALWAYS condition */
}


IP *lookupip(char *name, int insertf)
{
  int h = name[0];
  IP *r = iphashtab[h];
  while(r)
    {
      char *n = name;
      char *p = r->name;
      if (*n == *p)
	{
	  while (*n && *p)
	    {
	      if (*n == *p)
		{
		  n++; p++;
		  continue;
		}
	      if (*p == '*' && iscondition(n) != 14) 
		{
		  n+= 2; p++;
		  continue;
		}
	      if (*p == '*')
		{
		  p++;
		  continue;
		}
	      break;
	    }
	  if (*n == 0 && (*p == 0 || (*p=='*' && p[1] == 0))) break;
	}
      r = r-> next;
    }
  if (r && insertf) printf("opcode %s dmto\n", name); 
  if (r) return r;
  if (insertf)
    {
 
      r = NEW(IP);
      r->name = name;
      r->next = iphashtab[h];
      iphashtab[h] = r;
    }
  return r;
}

/* routine looks up nemonics  */
void lok()
{
  IP *ip;
  char nem [132];
  int i = 0;
  char xch = toupper(inst[tab++]);


  while (xch == ' ') xch = toupper(inst[tab++]);
  if (xch == (char) 0) return;       /* blank line */
  if (xch == ';' || xch == '@')  return;           /* comment */
  while(isalpha(xch) || xch == '.')
    {
      nem[i++] = xch;
      xch = toupper(inst[tab++]);
    }
  nem[i] = (char) 0;
  conditional_eval_code = 0xE0000000;
  ip = lookupip(nem, 0);
  
  if (ip == NULL) 
    {
      error("bad instruction ", strdup(nem));
      
      printf(" bad bog `%s' %s\n", nem, inst);
    }
  else
    {
      nextch(1);
      (ip->execute)(inst+tab-1, ip);
    }
}

/* specific prototypes */
int parse_immed();
int getpling();
void parse(PI *ps);

int parse_immed_or_reg()
{
  PI p;
  parse(&p);
  if (p.mod == m_reg) return (p.val<< 4) + 1;
  if (p.mod != m_immed) error("Immediate value expected", "");
  return (p.val << 3);
}

int readshift()
{
  inst[tab-1] = tolower(inst[tab-1]);
  inst[tab] = tolower(inst[tab]);
  inst[tab+1] = tolower(inst[tab+1]);

  if (!strncmp(inst+tab-1, "lsl", 3) || !strncmp(inst+tab-1, "asl", 3))
    {
      int v = 0;
      int d;
      nextch(1);      nextch(1);      nextch(1);
      d = parse_immed_or_reg();
      v = v | d;
      return v;
    }
  if (!strncmp(inst+tab-1, "lsr", 3))
    {
      int v = 1 << 1;
      int d;
      nextch(1);      nextch(1);      nextch(1);
      d = parse_immed_or_reg();
      v = v | d;
      return v;
    }
  if (!strncmp(inst+tab-1, "asr", 3))
    {
      int v = 2 << 1;
      int d;
      nextch(1);      nextch(1);      nextch(1);
      d = parse_immed_or_reg();
      v = v | d ;
      return v;
    }
  
  if (!strncmp(inst+tab-1, "rot", 3))
    {
      int v = 3 << 1;
      int d;
      nextch(1);      nextch(1);      nextch(1);
      d = parse_immed_or_reg();
      v = v | d;
      return v;
    }
  
  else error("bad shift %s", strdup(tab+inst-1));
  return 0;
}

void parse(PI *ps)
{
  /*  char *parsestart = inst+tab-1; */

  ps->mod = -1;  /* invalid value */

  /* find bit addressable things, including register names and register bits 
   */
  if (isalpha(ech))
    {
      int starttab = tab;
      char token[132];
      int i = 0;
      while (isalnum(ech) || (ech=='_') )       /* symbol look up */
      {
	token[i++] = ech; 
	nextch(0);
      }
      token[i++] = (char) 0;
      if (vlookup(token, 0))
	{
   
	  if (wordnode->type == GPREG)
	    {
	      ps->mod = m_reg;
	      ps->val = wordnode->value;
	      ignblank();
	      return;

	    }

	  else if (wordnode->type == SFR)
	    {
	      ps->val = wordnode->bshort;
	      ps->val1 = wordnode->blong;
	      ps->mod = m_sfr;
	      ignblank();
	      return;
	    }
	  /* otherwise, will lookup and find again in expression eval */  
	}
      if (0) printf("Not found %s\n", token);
      tab = starttab-1;
      nextch(1);
      
    }

  /* find pre and post forms */
  if (ech == '[')
    {

      PI p1, p2;
      nextch(1);
      parse (&p1);
      if (p1.mod != m_reg) 
	{
	  printf("DBX %i %c\n", p1.mod, ech);
	  error("base reg expected", "");
	}
      ps->val = p1.val;
      
      if (ech == ']')  /* early close means post */
	{
	  nextch(1);
	  ps->mod = m_post;
	  ps->offset = 0;
	}
      else ps->mod = m_pre;

      if (ech == ',') 
	{
	  chcomma();
	  if (ech == '-')
	    {
	      ps->sign = 1;
	      nextch(1);
	    }
	  else ps->sign = 0;
     

	  parse (&p2);
	  if (p2.mod == m_immed)
	    {
	      ps->val1 = -1;
	      ps->offset = p2.val;
	    }
	  else if (p2.mod == m_reg)
	    {
	      ps->val1 = p2.val;
	    }

	  ignblank();
	  if (ech == ',')
	    {
	      nextch(1);
	      ps->shift = readshift();
	    }
	  else ps->shift = 0;
	  
	  if (ech == ']')
	    {
	      nextch(1);
	    }
	  else if (ps->mod == m_pre) 
	    {
	      printf("Char .%c.\n", ech);
	      error("missing close bracket", "");
	    }
	}
      else
	{
	  ps->offset = 0;
	  ps->val1 = -1;
	}
      return;
    }

  if (ech == '#') /* find immediate addressing */
    {
      STNODE *poi;
      ps->mod = m_immed;
      ps->val = eval(&poi);
      return;
    }


  /* otherwise eval it */
  {
    STNODE *poi;
    ps ->mod = m_abs;
    ps ->val = eval(&poi);
   return;
  }
}

int parsereg()
{
  PI p;
  parse(&p);
  if (p.mod != m_reg) error("Register expected", "");
  return p.val;
}


int parse_immed()
{
  PI p;
  parse(&p);
  if (p.mod != m_immed) error("Immediate value expected", "");
  return p.val;
}


int applycond(int op)
{
  return op | conditional_eval_code;
}


/* rotate left till packed into bottom byte */
unsigned int pack_arm_imed(unsigned int vin)
{
  unsigned int v = vin;
  int r;
  char line[132];
  for (r=0; r<16; r++)
    {
      if (v < 256) 
	{
	  return (v | (r << 8));
	}

      v = (v << 2) | ((v >> 30) & 3);
    }
  sprintf(line, "%08X", vin);
  error("Impossible immediate constant %s", strdup(line));
  return 0;
}


void i_origin(char *v, IP *self)
{
  STNODE *p;
  extern void org(int);
  int o = eval (&p);
  org(o);
}

void i_padorigin(char *v, IP *self)
{
  STNODE *p;
  extern void padorg(int);
  int o = eval (&p);
  padorg(o);
}

void i_movement(char *v, IP *self)
{
  int op = self->i;
  int rd = parsereg();
  PI p;

  chcomma();
  parse(&p);

  op |= (rd << 12);

  if (p.mod == m_pre || p.mod == m_post)
    {
      op |= (p.val << 16); /* insert base reg */

      if (p.val1 < 0) /* use offset */
	{
	  int offset = p.offset;
	  if (p.sign) offset = -offset;
	  if (abs(offset) >= 4096) error("index out of range", "");	
	  op |= abs(offset) & 0xFFF;
	  if (offset > 0) op |= (1<<23); /* up down bit */

	}
      else
	{
	  op |= (1<<25) | (p.shift << 4) | (p.val1 << 0);
	  if (p.sign==0) op |= (1<<23); /* up down bit - set for up */ 
	}
      if (p.mod == m_pre) op |= (1<<24);  /* Set pre bit if pre */
      if (p.mod == m_post) op |= (1<<21);  /* Always write back for post indexed */
    }
  else if (p.mod == m_abs)
    {
      int offset = p.val - (baddr[output] + 8);
      if (offset > 0) op |= (1<<23); /* up down bit */ 
      offset = abs(offset);
      if (offset >= 4096) error("too far to become pc relative", "");
      op |=   (15 << 16) | (offset & 0xFFF);
      op |= (1<<24);  /* Set pre bit  */
    }
  else 
    {
      printf("%i\n", p.mod);
      error("bad movement addressing mode", "");
    }
  if (getpling()) op |= (1<<21);

  op = applycond(op);
  objword(op);
}

void i_alu(char *v, IP *self)
{
  int op = self->i;
  int rd;
  int rn;
  int sflag = 0;
  PI p;

  if (((self->i >> 23) & 3) == 2)  sflag = 1;  /* CMP, TST TEQ CMN have it default */


  if (sflag == 0)
    {
      rd = parsereg();
      chcomma();
    }
  else rd = 0;


  rn = parsereg();
  op = op + (rd << 12) + (rn << 16);


  chcomma();
  parse(&p);

  

  if (p.mod == m_reg)
    {
      op = op + p.val;
      if (ech != (char) 0 && ech != '@' && ech != ';') /* not eol */
	{
	  int shift;
	  chcomma();
	  shift = readshift();
	  op |= shift << 4;
	}
    }
  else if (p.mod == m_immed)
    {
      op = op + (1<<25) + pack_arm_imed(p.val);
    }

  else error("bad addressing mode", "");


  op = applycond(op);
  objword(op);
}


void i_mov(char *v, IP *self)
{
  int op = self->i;
  int rd = parsereg();

  PI p;
  op = op + (rd << 12);
  chcomma();
  parse(&p);
  ignblank();

  if (ech != (char) 0 && ech != '@' && ech != ';') /* not eol */
    {
      int shift;
      chcomma();
      shift = readshift();
      op |= shift << 4;
    }

  if (p.mod == m_reg)
    {
      op = op + p.val;
    }
  else if (p.mod == m_immed)
    {
      op = op + (1<<25) + pack_arm_imed(p.val);
    }

  else error("bad addressing mode", "");

  op = applycond(op);
  objword(op);
  /* */
}


void i_null(char *v, IP *self)
{
  /* */
}



/* uncond relative jump and sub jump */
void i_jmp(char *v, IP *self)
{
  STNODE *poi;
  int dest = eval(&poi);
  int o;
  int delta = (dest - (baddr[output] + 8)) >> 2;


  if (delta > (1<<24)-1 || delta < -(1<<24)) 
    {
      error("branch out of range", "");
    }

  o = self->i + (delta & 0xFFFFFF);
  
  o = applycond(o);  
  objword(o);
  return;
}

void i_tbsel(char *v, IP *self)
{
  output = self->i;
}

int getpling()
{
  if (ech == '!')
    {
      nextch(1);
      return 1;
    }
  return 0;
}


void i_multiply(char *v, IP *self)
{
  int rd, rm, rs, rn;
  int o = self->i;
  rd = parsereg();
  chcomma();


  rm = parsereg();
  chcomma();


  rs = parsereg();
  
  
  if (o & (1<<21))
    {
      chcomma();  
      rn = parsereg();
    }
  else rn = 0;

  o |= (rd <<16) | (rn << 12) | (rs<<8) | (rm);

  o = applycond(o);
  objword(o);

}

void i_multiple(char *v, IP *self)
{
  int rn;
  int o = self->i;
  int pling;

  rn = parsereg();
  pling = getpling();
  chcomma();

  o |= (rn << 16);
  if (pling) o |= (1<<21);

  if (ech == '{')
    {
      nextch(1);
      while (1)
	{
	  int x = parsereg();
	  if (ech == '-')
	    {
	      int y;
	      nextch(1);
	      y = parsereg();
	      while (1)
		{
		  o |= (1 << x);
		  if (x == y) break;
		  if (x < y) x++; else x--;
		}
	    }
	  else o |= (1 << x);
	  if (ech =='}') break;
	  chcomma();
	}
    }
  else error("register list needed in { }", "");

  if (ech == '^') 
    {
      nextch(1);
      o |= (1<<22);
    }
  o = applycond(o);
  objword(o);
}

void i_defs(char *v, IP *self)
{
  STNODE *poi;
  int a = eval(&poi);
  while (a > 0)
    {
      objword(0);
      a -= 4;
    }
}

void i_adr(char *v, IP *self)
{
  int rd;
  STNODE *poi;
  int dest, delta;
  int o;
  rd = parsereg();
  chcomma();
  dest = eval(&poi);
  delta = dest - (baddr[output] + 8);
  
  if (delta < 0) 
    {
      delta = -delta;
    }
  else o |= (1<<23);
  /* unconditional add instruction to pc, result in rd */
  o = 0xE0000000 | 0x00800000 | (rd << 12) | (15 << 16);

  o |= pack_arm_imed(delta)  | (1<<25);
 
  objword(o);
}



void i_defw(char *v, IP *self)
{
  STNODE *poi;
  int va = eval(&poi);
  objword(va);
}

void i_align(char *v, IP *self)
{
  STNODE *poi;
  /*int va = eval(&poi);
   */
  int d = baddr[output]  & 3;
  if (d) while (d++ < 4) objbyte(0);
}

void i_implied(char *v, IP *self)
{
  objword(self->i);
}

void i_defb(char *v, IP *self)
{
  while(1)
    {
      if (ech == (char) 0x22 || ech == '\'')
	{
	  char startchar = ech;
	  nextch(0);
	  while (ech!= startchar)
	    {
	      if (ech == '\\')
		{
		  int r = 0;
		  nextch(0);
		  while (isdigit(ech))
		    {
		      r = (r << 3) + ech - '0';
		      nextch(0);
		    }
		  objbyte(r);
		}
	      else
		{
		  objbyte(ech);
		  nextch(0);
		}
	    }
	  nextch(1);
	}
      else
	{
	  STNODE *poi;
	  int va = eval(&poi);
	  objbyte(va);
	}
      if (ech == 0 || ech == ';' || ech == '@') return;
      chcomma();
    }
}

void install_ii(char *name, void (*ex)(), int i)
{
  IP *r = lookupip(name, 1);
  r->name = name;
  r->execute = ex;
  r->i = i;
}

void install_cond(int code, char *name)
{
  IP *r = lookupip(name, 1);
  r->name = name;
  r->execute = 0;
  r->i = code << 28;
}


void install_r(char *name, int val, int type)
{
  vlookup(name, 1);
  wordnode -> type = type;
  wordnode -> value = val;
}
   
/* define a special function register - self indicates bit addressable */
void i_regdecl(char *v, IP *self)
{
  PI pi;
  parse(&pi);

  if (pi.mod != m_reg) error("R0-R15 needed", "");
  if (lhsnode)
    {
      lhsnode ->value = pi.val;
      lhsnode ->type = GPREG;
      lhsnode->equatef = 1;
    }
  else error("missing reg in lab field", "");


}

/*
 * equate directive 
 */
void i_equ(char *v, IP *self)
{
  STNODE *ps0;
  int vv = eval(&ps0);

   if (lhsnode)
    {
      if (pass == 1 && lhsnode ->equatef) lhsnode->type = DMTO;
      else
	{
	  lhsnode ->value = vv;
	  lhsnode ->equatef = 1;
	  lhsnode ->type = ABSOLUTE;
	}
    }
  else error("missing label", "");


}



void i_copro(char *v, IP *self)
{
  STNODE *poi;
  int p, e1, rd, cn, cm, e2;
  int o = self ->i;

  p = eval(&poi);  /* 1 -processor */
  chcomma();
  if (1) 
    {
      e1 = eval(&poi);  /* 2 - not used - user always puts to zero */
      chcomma();
    }
  else e1 = 0;
  rd = parsereg();  /* 3 - arm reg */
  chcomma();
  cn = eval(&poi);  /* 4 - copro reg */
  chcomma();
  cm = eval(&poi);  /* 5 - cm value */
  if (ech == ',')
    {
      chcomma();
      e2 = eval(&poi);  /* 6 - optional OPC_2 code */
    }
  else e2 = 0;

  o |= (e1 << 21) | (cn << 16) | (rd << 12) | (p << 8) | (e2 << 5) | cm; 
  o = applycond(o);
  objword(o);
      
}



/*
 *
 * mrs Rd,CPSR or SPSR     ; Read from status to gpr
 * msr CPSR or SPSR,Rd     ; Write to status
 * msr CPSR or SPSR,#imed
 */  
int pp_spsr()
{
  int r = 0, f = 0;
  if (!strncmp(inst+tab-1, "CPSR", 4))
    {
      r = 1;
    }
  else if (!strncmp(inst+tab-1, "SPSR", 4))
    {
      r = 2;
    }
  
  if (r) 
    {
      nextch(1);      nextch(1);      nextch(1);      nextch(1);
    }
  else error("CPSR OR SPSR expected", "");

  if (!strncmp(inst+tab-1, "_all", 4)) f = 1;
  if (!strncmp(inst+tab-1, "_flg", 4)) f = 2;

  if (f) 
    {
      nextch(1);      nextch(1);      nextch(1);      nextch(1);
    }
  
  
  return r + 16 * f;
}



void i_mrs(char *v, IP *self)
{
  int rd;
  int o = self ->i;
  rd = parsereg();
  chcomma();
  if ((pp_spsr() & 3) == 2) o |= (1<<22); /* Set bit 22 if saved SPSR */
  o |= (rd << 12);
  o = applycond(o);
  objword(o);
}

void i_msr(char *v, IP *self)
{
  PI p;
  int o;
  int vv = pp_spsr();

  if ((vv & 3) == 2) o |= (1<<22);
  chcomma();
  parse(&p);
  
  if ((vv & 0xF0) == 0x10)
    {
      if (p.mod == m_reg)
	{
	  o = 0x0129F000 | p.val;
	}
      else error("register needed if all bits to be set", "");
    }
  else  if ((vv & 0xF0) == 0x20)
    {
      o = 0x0128F000;
      if (p.mod == m_reg)
	{
	  o |=  p.val;
	  if (ech != (char) 0 && ech != '@' && ech != ';') /* not eol */
	    {
	      int shift;
	      chcomma();
	      shift = readshift();
	      o |= shift << 4;
	    }
	}
      else if (p.mod == m_immed)
	{
	  o |= (1<<25) + pack_arm_imed(p.val);
	}
      else error("bad addressing mode", "");
    }
  else error("must give _flg or _all qualifier", "");

  o = applycond(o);
  objword(o);
}


void i_copy(char *v, IP *self)
{
  extern void gas_copy(int f);
  if (ech == '0') gas_copy(0);
  else if (ech == '1') gas_copy(1);
  else error("bad copy arg", "");
}


void i_get(char *v, IP *self)
{
  extern void gas_get();
  gas_get();
}


void initspecific()
{
  int i;
  for (i=0;i<256;i++) iphashtab[i] = NULL;

  /* mrc reads coprocessor to gp reg - set bit 20 */
  install_ii("MRC*", i_copro, 0x0E100010);
  /* mcr write to coprocessor */
  install_ii("MCR*", i_copro, 0x0E000010); 
  
  install_ii("MSR*", i_msr, 0);
  install_ii("MRS*", i_mrs, 0x010F0000); /* fixed jul 98 */

  install_ii("EQU", i_equ, 0);
  install_ii(".EQU", i_equ, 0);

  install_ii("AND*", i_alu,  0x00000000);
  install_ii("AND*S", i_alu, 0x00100000);
  install_ii("EOR*", i_alu,  0x00200000);
  install_ii("EOR*S", i_alu, 0x00300000);

  install_ii("SUB*", i_alu,  0x00400000);
  install_ii("SUB*S", i_alu, 0x00500000);

  install_ii("RSB*", i_alu,  0x00600000);
  install_ii("RSB*S", i_alu, 0x00700000);

  install_ii("ADD*", i_alu,   0x00800000);
  install_ii("ADD*S", i_alu,  0x00900000);

  install_ii("ADC*", i_alu,   0x00A00000);
  install_ii("ADC*S", i_alu,  0x00B00000);

  install_ii("SBC*", i_alu,   0x00C00000);
  install_ii("SBC*S", i_alu,  0x00D00000);

  install_ii("RSC*", i_alu,   0x00E00000);
  install_ii("RSC*S", i_alu,  0x00F00000);

  install_ii("TST*", i_alu,   0x01100000);
  install_ii("TEQ*", i_alu,   0x01300000);
  install_ii("CMP*", i_alu,   0x01500000);
  install_ii("CMN*", i_alu,   0x017100000);

  install_ii("OR*", i_alu,    0x01800000);
  install_ii("OR*S", i_alu,   0x01900000);

  install_ii("ORR*", i_alu,    0x01800000);
  install_ii("ORR*S", i_alu,   0x01900000);

  install_ii("MOV*", i_mov, 0x01A00000);
  install_ii("MOV*S", i_mov, 0x01B00000);

  install_ii("BIC*", i_alu,   0x01C00000);
  install_ii("BIC*S", i_alu,  0x01D00000);

  install_ii("MVN*", i_mov, 0x01E00000);
  install_ii("MVN*S", i_mov, 0x01F00000);


  install_ii("BL*", i_jmp, 0x0B000000);
  install_ii("B*", i_jmp, 0x0A000000);
 
  install_ii("MLA*", i_multiply, 0x00200090);
  install_ii("MUL*", i_multiply, 0x00000090);
  install_ii("MLA*S", i_multiply, 0x00300090);
  install_ii("MUL*S", i_multiply, 0x00100090);


  install_ii("LDR*B", i_movement, 0x04500000);
  install_ii("STR*B", i_movement, 0x04400000);
  install_ii("LDR*", i_movement, 0x04100000);
  install_ii("STR*", i_movement, 0x04000000);

  install_ii("LDM*ED", i_multiple, 0x09900000); /* L, U and Post  */
  install_ii("LDM*IB", i_multiple, 0x09900000); /* L, U and Post  */

  install_ii("LDM*IA", i_multiple, 0x08900000); /* L and U */
  install_ii("LDM*FD", i_multiple, 0x08900000); /* L and U */

  install_ii("LDM*EA", i_multiple, 0x09100000); /* L and Post  */
  install_ii("LDM*DB", i_multiple, 0x09100000); /* L and Post  */

  install_ii("LDM*DA", i_multiple, 0x08100000); /* L   */
  install_ii("LDM*FA", i_multiple, 0x08100000); /* L   */

  install_ii("STM*FA", i_multiple, 0x09800000); /* P and U */
  install_ii("STM*IB", i_multiple, 0x09800000); /* P and U */

  install_ii("STM*IA", i_multiple, 0x08800000); /* just U */
  install_ii("STM*EA", i_multiple, 0x08800000); /* just U */

  install_ii("STM*FD", i_multiple, 0x09000000); /* just Post */
  install_ii("STM*DB", i_multiple, 0x09000000); /* just Post */

  install_ii("STM*ED", i_multiple, 0x08000000); /* none */
  install_ii("STM*DA", i_multiple, 0x08000000); /* none */

  install_ii("NOP", i_implied, 0xE1A00000);
  install_ii("ADR", i_adr, 0);

  install_ii("DC", i_defw, 0);
  install_ii(".WORD", i_defw, 0);
  install_ii(".DEFW", i_defw, 0);
  install_ii(".BYTE", i_defb, 0);
  install_ii(".DATA", i_defb, 0);
  install_ii(".SPACE", i_defs, 0);
  install_ii(".DEFS", i_defs, 0);
  install_ii(".ASCII", i_defb, 0);

  install_ii(".GLOBAL", i_null, 0);
  install_ii(".EXTERN", i_null, 0);
  install_ii(".END", i_null, 0);
  install_ii(".STABS", i_null, 0);
  install_ii(".STABN", i_null, 0);
  install_ii(".STABD", i_null, 0);
  install_ii(".ALIGN", i_align, 0);
  install_ii(".COMM", i_null, 0);
  install_ii(".REQ", i_regdecl, 0);
  install_ii(".TEXT", i_tbsel, O_TEXT);
  install_ii(".BSS", i_tbsel, O_BSS);

  install_ii("INCLUDE", i_get, 0);
  install_ii(".COPY", i_copy, 0);
  install_ii("COPY", i_copy, 0);
  /*
   *  install_i("GET", i_get);
   */
  install_ii("ORG", i_origin, 0);
  install_ii("PADORG", i_padorigin, 0);

  /*
  install_cond(0, "EQ");
  install_cond(1, "NE");
  install_cond(2, "CS");
  install_cond(3, "CC");
  install_cond(4, "MI");
  install_cond(5, "PL");
  install_cond(6, "VS");
  install_cond(7, "VC");
  install_cond(8, "HI");
  install_cond(9, "LS");
  install_cond(10, "GE");
  install_cond(11, "LT");
  install_cond(12, "GT");
  install_cond(13, "LE");
  */

  install_r("r0", 0, GPREG);
  install_r("r1", 1, GPREG);
  install_r("r2", 2, GPREG);
  install_r("r3", 3, GPREG);
  install_r("r4", 4, GPREG);
  install_r("r5", 5, GPREG);
  install_r("r6", 6, GPREG);
  install_r("r7", 7, GPREG);
  install_r("r8", 8, GPREG);
  install_r("r9", 9, GPREG);
  install_r("r10", 10, GPREG);
  install_r("r11", 11, GPREG);
  install_r("r12", 12, GPREG);
  install_r("r13", 13, GPREG);
  install_r("r14", 14, GPREG);
  install_r("r15", 15, GPREG);

  install_r("R0", 0, GPREG);
  install_r("R1", 1, GPREG);
  install_r("R2", 2, GPREG);
  install_r("R3", 3, GPREG);
  install_r("R4", 4, GPREG);
  install_r("R5", 5, GPREG);
  install_r("R6", 6, GPREG);
  install_r("R7", 7, GPREG);
  install_r("R8", 8, GPREG);
  install_r("R9", 9, GPREG);
  install_r("R10", 10, GPREG);
  install_r("R11", 11, GPREG);
  install_r("R12", 12, GPREG);
  install_r("R13", 13, GPREG);
  install_r("R14", 14, GPREG);
  install_r("R15", 15, GPREG);
  install_r("PC", 15, GPREG);
  install_r("pc", 15, GPREG);

}


/*
 *   ADR r12, blah
 *
 * actually assembles as
 *
 *  ADD r12, pc, #blah - * - 8
 *or
 *   SUB r12, pc, #* + 8 - blah
 */

/* end of cbg carmas.c */

 
