/*
 * emulator.c   ARM emulator.          * TTRX ARMBOS MAR96
 *
 *
 *
 * (C) 1990-1996 Tenison Technology 
 * info@tenisontech.com.  Fax +44 1223 565917
 *
 * No responsibility can be accepted for loss or consequential loss.
 * Not for commercial use without license.  
 *
 * Software Excellence from Tenison Technology.
 *
 * (C) 1995 Tenison Technology.
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

/* Change endianness as needed */
#define HTOAENDIAN(X) ((byteswapf) ? X: byteswap(X)) 
#define ATOHENDIAN(X) ((byteswapf) ? X: byteswap(X)) 

#define START_STACK 0x4000
#define LOW_ADDRESS 0x80000
#define HIGH_ADDRESS 0x9FFFF
#define MEMSIZE (HIGH_ADDRESS-LOW_ADDRESS+1)
#define ENTRY_POINT LOW_ADDRESS
void find_a_symbol(int v, int lf);

#define bptype_bp  0  /* break and stop */
#define bptype_bpt 1  /* show regs then continue */
#define bptype_bpl 2  /* break and start listing each step */
int arm7_total_cycles = 0;
int instruction_limit = 10000000;
int entry_point = ENTRY_POINT;
int nomemreportf = 0;
int bpcont = 0;
int running = 1;
int hitpoint = 0;
int listeachstepf = 0;
int breakpoint = 0;
int byteswapf = 1;
int disasf = 0;
int ihexf = 1;
int points = 4;
typedef unsigned char uchar;
int verbosef = 0;

char *symbols, *inputfile;
char *shiftstrings[4] = { "ASL",  "LSR",  "ASR",  "ROR" };

void myadis(unsigned int x, int a)
{
  char *conde;  
  printf("%08X %08X ", a, x);
  switch(x >> 28)
    {	
    case 0: conde = "eq"; break;
    case 1: conde = "ne"; break;
    case 2: conde = "c?"; break;
    case 3: conde = "c?"; break;
    case 4: conde = "mi"; break;
    case 5: conde = "pl"; break;
    case 6: conde = "vs"; break;
    case 7: conde = "vc"; break;
    case 8: conde = "hi"; break;
    case 9: conde = "ls"; break;
    case 10: conde = "ge"; break;
    case 11: conde = "lt"; break;
    case 12: conde = "gt"; break;
    case 13: conde = "le"; break;
    case 14: conde = ""; break;
    case 15: conde = "nv"; break;
    }
  
  if ((x & 0x0F000000)==0x0F000000)
    {
      printf("swi%s 0x%x\n", conde, x & 0xFFFFFF);
    }
  else if ((x & 0x0F000000)==0x0E000000)
    {
      int cn =  (x >> 16) & 15;
      int rd =  (x >> 12) & 15;
      int p =  (x >> 8) & 15;
      int e2 =  (x >> 5) & 7;
      int cm = x & 15;
      char *op = (x & (1<<20)) ? "mrc": "mcr";
      printf(" %s  %i,R%i,CP%i,%i,%i", op, p, rd, cn, cm, e2);
    }
  else if ((x & 0x0E000000)==0x0A000000)
    {  
      char *s;
      int o = (x & 0x00FFffFF);
      o = o << 2;
      if (o & 0x02000000) o = o | 0xFC000000; 
      if (x & 0x01000000) s = "bl"; else s = "b";
      printf("%s%s 0x%08X", s, conde, a + o + 8);
      find_a_symbol(a+o+8, 0);
      printf("\n");
    }
  else if ((x & 0x0E000000)==0x08000000)
    {  
      char d, w, *s;
      unsigned int i;
      if (x & 0x00100000) s = "ldm"; else s = "stm";

      if (x & 0x00800000) d='+'; else d='-';
      if (x & 0x00200000) w='!'; else w=' ';
      printf("%s%s %c%c ", s, conde, d, w);
      printf("  R%i,{ ", (x >>16) & 15);
      for (i=0;i<16; i++) if (x & (1<<i)) printf("R%i,", i);
      printf("}\n");
      
    }
  else if ((x & 0x0D900000)==0x01000000)
    {
      printf("move sr\n");
    }
  else if ((x & 0x0FC000F0)==0x00000090)
    {
      printf("mul%s%c%c R%i,R%i,R%i,R%i\n", conde,
	     (x & 0x00200000) ? 'a':' ',
	     (x & 0x00100000) ? 'S':' ',
	     ((x>>16) & 15),
	     ((x>>12) & 15),
	     ((x>>8) & 15),
	     ((x>>0) & 15));
    }
  else if ((x & 0x0C000000)==0x00000000)
    {
      char * s;
      unsigned int op = (x >> 21) & 15;
      switch(op)
	{
	case 0 : s = "and"; break;
	case 1 : s = "eor"; break;
	case 2 : s = "sub"; break;
	case 3 : s = "rsb"; break;
	case 4 : s = "add"; break;
	case 5 : s = "adc"; break;
	case 6 : s = "sbc"; break;
	case 7 : s = "rsc"; break;
	case 8 : s = "tst"; break;
	case 9 : s = "teq"; break;
	case 10 : s = "cmp"; break;
	case 11 : s = "cmn"; break;
	case 12 : s = "orr"; break;
	case 13 : s = "mov"; break;
	case 14 : s = "bic"; break;
	case 15 : s = "mvn"; break;
	}
      
      /* Print nemonic and Rd if used */
      if (op >=8 && op <=11) printf ("%s%s ", s, conde);
      else printf ("%s%s%c R%i,", s, conde, (x & (1<<20)) ? 's':' ',  (x >> 12) & 15);
      
      /* Print Op2 if used */
      if (op != 13 && op != 15) printf("R%i,", (x >>16) & 15);
      
      /* Print Op1 */
      if (x & 0x02000000) /* Immediate */
	{
	  unsigned int v = (x & 0xff);
	  unsigned int shift = (x & 0xf00) >> 8;
	  while (shift-- > 0) v = (v >> 2) | (v << 30);
	  printf("#0x%x", v);
	  if (v >= 32 && v < 127) printf(" '%c'", v);
	}
      else
	{
	  int shift = x & 0xFF0;
	  printf("R%i", x & 15);
	  if (shift) 
	    {
	      shift = shift >> 4;
	      printf(",%s ", shiftstrings[(shift & 6) >> 1]);
	      if (shift & 1) printf("R%i", shift >> 4);
	      else printf("#%i", shift >> 3);
	    }
	  
	}
      printf("\n");
    }
  else if ((x & 0x0c000000)==0x04000000 || (x & 0x0E000000) == 0x0C000000)
    {
      char *s;
      if ((x & 0x0E000000) == 0x0C000000) printf(" copro ");
      if (x & 0x00100000) s = "ldr"; else s = "str";
      printf("%s%s%c %c R%i,", 
	     s, 
	     conde,
	     ((x & 0x00400000) ? 'b':' '), 
	     ((x & 0x00200000) ? '!':' '), 
	     ((x >> 12) & 15));

      if ((x & 0x02000000)==0)
	{
	  int o = x & 0xFFF;
	  unsigned int b = (x>>16) & 15;
	  if ((x & 0x00800000)==0) o = -o;
	  if (x & 0x01000000)
	    {
	      if (o == 0) printf("[R%i]", b);
	      else printf("[R%i,#%i]", b, o);
	      if (x & 0x00200000) printf("!");
	    }
	  else
	    {
	      if (o == 0) printf("[r%i]", b);
	      else printf("[r%i],#%i", b, o);
	      if (x & 0x00200000) printf("!");
	    }
	  if (b == 15) printf("  (0x%08x)", a + o + 8);
	
	}
      else
	{
	  unsigned int shift = (x >> 4) & 0xff;
	  unsigned int r = x & 15;
	  unsigned int b = (x>>16) & 15;
	  char d;
	  if ((x & 0x00800000)==0) d ='-'; else d = '+';
	  if (x & 0x02000000)
	    {
	      if (shift)
		{
		  shift = shift >> 4;
		  printf("[r%i,%cr%i BAD%i,", b, d, r, shift);
		  printf(",%s ", shiftstrings[(shift & 6) >> 1]);
		  if (shift & 1) printf("R%i]", shift >> 4);
		  else printf("#%i]", shift >> 3);
		}
	      else  printf("[R%i,%cR%i]", b, d, r);
	      if (x & 0x00200000) printf("!");
	    }
	  else
	    {
	     
	      if (shift)  printf("[R%i],%cR%i, shift=%i", b, d, r, shift);
	      else  printf("[R%i],%cR%i", b, d, r);
	      if (x & 0x00200000) printf("!");
	    }
	}
      printf("\n");
    }
  else printf(" ??? \n");
}


unsigned int test_data[] = 
{
  0x8C02A0E3, 0x00F0A0E1, 0x0000A0E3, 0x00B0A0E3,
  0x00A0A0E3, 0x84D09FE5, 0x410200EB, 0xFEFFFFEA,
  0x00000FE1, 0xC000C0E3, 0x00F029E1, 0x0EF0A0E1,
  0x6C009FE5, 0x6C109FE5, 0x000081E5, 0x010000EF,
  0xFEFFFFEA, 0xBC0900EB, 0x200900EB, 0x120000EA,
  0x00000FE1, 0x1F00C0E3, 0x100080E3, 0x00F029E1,
  0x00B0A0E3, 0x00A0A0E3, 0x100000EF, 0x01D0A0E1,
  0x7F1CA0E3, 0x00D081E5, 0x4800A0E3, 0x000000EF,
  0x7700A0E3, 0x000000EF, 0x0D00A0E3, 0x000000EF,
  0x0A00A0E3, 0x000000EF, 0x0C0900EB, 0xFEFFFFEA,
  0xF0FF0000, 0x440000C0, 0x48080000, 0x0E10A0E1,
  0x0D20A0E1, 0x00000FE1, 0x1F00C0E3, 0x130080E3,
  0x00F029E1, 0x02D0A0E1, 0x01F0A0E1, 0x000000EF,
  0x0EF0A0E1, 0x040000EF, 0x0EF0A0E1, 0x020000EF,
  0x0EF0A0E1, 0x00000FE1, 0x0EF0A0E1, 0x160000EF,
  0x0000E0E3, 0x0EF0A0E1, 0x60D09FE5, 0x00008DE5,
  0x04108DE5, 0x08208DE5, 0x0C308DE5, 0x8024A0E3,
  0x140092E5, 0x040010E2, 0x0B00000A, 0x40309FE5,
  0x000093E5, 0x3C109FE5, 0x010080E0, 0x0C1092E5,
  0x0010C0E5, 0x000093E5, 0x010080E2, 0x000BA0E1 
};		


unsigned int byteswap(unsigned x)
{
  return (x << 24) + ((x<<8) & 0x00FF0000) + ((x >> 8) & 0x0000FF00) + (x >> 24);
}


static unsigned char memory[MEMSIZE];
static int gpregs[16];
static int zf, nf, cf, vf;
static int mem_reads, mem_writes, steps = 0;
FILE *logfile;

void coredump()
{
  int i;
  int *m = (int *) (&memory[0]);
  
  for (i=0;i<=MEMSIZE>>2; i+=4)
    {
      int j;
      fprintf(logfile, "%08X:  %08X %08X %08X %08X  ", i<<2, m[i], m[i+1], m[i+2], m[i+3]);
      for (j=0;j<16;j++)
	{
	  putc((isprint(memory[i*4+j])) ? (memory[i*4+j]): ' ', logfile); 

	}
      fprintf(logfile, "\n");
    }
}

void showregs()
{
  printf("   R0=%08X  R1=%08X  R2=%08X  R3=%08X\n", gpregs[0], gpregs[1], gpregs[2], gpregs[3]);
  printf("   R4=%08X  R5=%08X  R6=%08X  R7=%08X\n", gpregs[4], gpregs[5], gpregs[6], gpregs[7]);
  printf("   R8=%08X  R9=%08X  Ra=%08X  Rb=%08X  Z N C V\n", gpregs[8], gpregs[9], gpregs[10], gpregs[11]);
  printf("   Rc=%08X  Rd=%08X  lr=%08X  pc=%08X  %i %i %i %i\n", gpregs[12], gpregs[13], gpregs[14], gpregs[15], zf, nf, cf, vf);

}

/*
 * Contact TenisonTech for the correct version of this code w.r.t the V flag.
 */
void update_zn_flags(unsigned int a)
{
  zf = (a == 0) ? 1 : 0;
  nf = (a >> 31);
}

int apply_shift(unsigned int d, int shift)
{
  int type, amount;
  if (shift & 1) amount = gpregs[shift >> 4];
  else amount = (shift >> 3);
  type = shift & 6;
  if (type == 0) /* logical left */
    {
      while (amount-- > 0) d = d << 1;
      return d;
    }

  else if (type == 2) /* logical right */
    {
      while (amount-- > 0) d = d >> 1;
      return d;
    }

  else if (type == 4) /* arith right */
    {
      int v = d;
      while (amount-- > 0) v = v >> 1;
      return v;
    }

  else if (type == 6) /* rotate right */
    {
      if (d == 0) { printf("EXT not supported\n"); exit(1);}
      while (d-- > 0) d = (d >> 1) | (d << 31);  
      return d;
    }
  else return 0;
}
  

/*
 * Return 1 if memory address is valid
 */
int check_addr(int a)
{
  if ( a < LOW_ADDRESS || a >= HIGH_ADDRESS)
    {
      if (nomemreportf ==0)
	{
	  printf("Emulated memory error, access address %x\n", a);
	  printf("Fault PC = %x\n", gpregs[15]-8);
	}
      if (0)
	{
	  running = 0;
	}
      else return 0;
    }
  else return 1;
}


int memory_read(int a, int bytef)
{
  mem_reads += 1;
  if (check_addr(a))
    {
      if (bytef) return memory[a-LOW_ADDRESS];
      else 
	{
	  int r = HTOAENDIAN(*((int *)(memory+a-LOW_ADDRESS)));
	  return r;
	}
    }
  return 0;
}


void emulate_putchar()
{
  putc(gpregs[0], logfile);
}

void emulate_getchar()
{
  if (inputfile == NULL || *inputfile == 0)
    {
      printf("***End of inputfile reached\n");
      fprintf(logfile, "\n\n***End of inputfile reached\n");
      running = 0;
    }
  else 
    {
      char c =  *inputfile ++; 
      printf("Read input char 0x%02X\n", c);
      gpregs[0] = c;
    }
}

void emulate_printf()
{
  int p = gpregs[0];
  char c = memory_read(p++, 1);
  char format[132];
  int i = 0;
  printf("Printf >");
  while (c && i <132)
    {
      format[i++] = c;
      c = memory_read(p++, 1);
    }
  format[i] = 0;
  printf(format, gpregs[1], gpregs[2], gpregs[3], gpregs[4], gpregs[5]); 
  fprintf(logfile, format, gpregs[1], gpregs[2], gpregs[3], gpregs[4], gpregs[5]); 
}

void memory_write(int a, int d, int bytef)
{
  if (check_addr(a))
    {
      mem_writes += 1;
      if (bytef) memory[a-LOW_ADDRESS] = d;
      else *((int *)(memory+a-LOW_ADDRESS)) = ATOHENDIAN(d);
    }
}



void step(int vb)
{
  int arm7_cycles = 1; /* Default to one and increase if needed */
  int a = gpregs[15]-8;
  int i = memory_read(a, 0);
  int d = (i & 0x0C000000);
  int cond = (i & 0xF0000000);
  int doitf;

  if (check_addr(a)==0) 
    {
      printf("Illegal PC instruction fetch address %08X\n", a);
      running = 0;
    }


  if (vb) myadis(i, gpregs[15]-8);
 

  if (cond == 0xE0000000) doitf = 1;
  else if (cond == 0x00000000) /* eq */
    {
      doitf = zf;  
    }

  else if (cond == 0x10000000) /* ne */
    {
      doitf = !zf;
    }
  else if (cond == 0x20000000) /* cs */
    {
      doitf = cf;  
    }

  else if (cond == 0x30000000) /* cc */
    {
      doitf = !cf;
    }

  else if (cond == 0x40000000) /* mi */
    {
      doitf = nf;
    }

  else if (cond == 0x50000000) /* pl */
    {
      doitf = !nf;
    }

  else if (cond == 0x60000000) /* vs */
    {
      doitf = vf;
    }


  else if (cond == 0x70000000) /* vc */
    {
      if (vf==0) doitf = 1;
      else doitf = 0;
    }

  else if (cond == 0x80000000) /* hi */
    {
      /* C set and Z clear */
      if (cf && zf==0) doitf = 1;
      else doitf = 0;
    }

  else if (cond == 0x90000000) /* ls */
    {
      /* C clear or Z set */
      if (zf || cf == 0) doitf = 1;
      else doitf = 0;
    }


  else if (cond == 0xA0000000) /* ge */
    {
      /* (N.V  or ~N.~V) */
      if (nf == vf) doitf = 1;
      else doitf = 0;
    }

  else if (cond == 0xB0000000) /* lt */
    {
      /* (N.~V  or N.~V) */
      if (nf != vf) doitf = 1;
      else doitf = 0;
    }

  else if (cond == 0xC0000000) /* gt */
    {
      /* Z clear and N==V */
      if (zf == 0 && nf == vf) doitf = 1;
      else doitf = 0;
    }

  else if (cond == 0xD0000000) /* le */
    {
      /* Z set or N^V */
      if (zf || nf != vf) doitf = 1;
      else doitf = 0;
    }


  else
    {
      if (i == 0xFFFF0011) 
	{
	  emulate_putchar();
	  doitf = 0;
	}

      else if (i == 0xFFFF0012) 
	{
	  emulate_getchar();
	  doitf = 0;
	}
    
      else if (i == 0xFFFF0001) 
	{
	  emulate_printf();
	  doitf = 0;
	}
      else
      {
	printf("Unsupported condition %x\n", cond);
	running = 0;
      }
    }


  if (doitf)
    {
      if (d == 0) /* data processing */
	{
	  if ((i & 0x0FC000F0) == 0x00000090) /* mul or mula */
	    {
	      int rd = (i >> 16) & 15;
	      int rn = (i >> 12) & 15;
	      int rs = (i >> 8) & 15;
	      int rm = (i >> 0) & 15;
	      int rsv = gpregs[rs];
	      int n = gpregs[rm] * rsv;
	      int m = 0;
	      int v = 1;
	      /* Determine extra cycles needed on ATM 7 for multiply */
	      while(m < 16 && ((unsigned) rsv) < ((unsigned) v))
		{
		  m += 1;
		  v = v << 2;
		}
	      arm7_cycles = m;

	      if (i & (1<<21)) gpregs[rd] = gpregs[rn] + n;
	      else gpregs[rd] = n;

	      if (i & (1<<20)) update_zn_flags(n);


	    }
	  else
	    {
	      int opcode = i & 0x01E00000;
	      int rd = (i >> 12) & 15;
	      int rn = (i >> 16) & 15;
	      
	      int op2;
	      if (i & (1<<25))
		{
		  /* immediate with shift */
		  int ashift = (i >> 8) & 15;
		  unsigned int o = i & 0xFF;
		  while (ashift-- > 0) o = (o >> 2) | (o << 30);
		  op2 = o;
		}
	      else 
		{
		  op2 = gpregs[i & 15];
		  op2 = apply_shift(op2, (i >> 4) & 0xFF);
		  if (i & (1<<4)) arm7_cycles += 1; /* Extra cycle when shift amount is in a register */
		}
	      
	      if (opcode == 0x00000000) /* AND */
		{
		  int n = gpregs[rn] & op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) update_zn_flags(n);
		}
	      
	      else if (opcode == 0x00200000) /* EOR */
		{
		  int n = gpregs[rn] ^ op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) update_zn_flags(n);
		}
	      
	      else if (opcode == 0x00400000) /* SUB */
		{
		  int op1 = gpregs[rn];
		  int n = op1 - op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) 
		    {
		      if (n < 0) cf = 0; else cf = 1;
		      vf = 0;
		      update_zn_flags(n);
		    }
		}
	      
	      else if (opcode == 0x00600000) /* RSB */
		{
		  int op1 = gpregs[rn];
		  int n = op2 - op1;
		  gpregs[rd] = n;	    
		  if (i & (1<<20))
		    {
		      update_zn_flags(n);
		      vf = 0;
		      if (n < 0) cf = 0; else cf = 1;
		    }
		}

	      else if (opcode == 0x00800000) /* ADD */
		{
		  int op1 = gpregs[rn];
		  int n = op1 + op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) 
		    {
		      unsigned int o1 = op1;
		      unsigned int o2 = -op2;
		      if (o1 > o2) cf = 1; else cf = 0;
		      update_zn_flags(n);
		    }
		}

	      else if (opcode == 0x01000000) /* TST */
		{
		  int op1 = gpregs[rn];
		  int n = op1 & op2;
		  if (i & (1<<20)) 
		    {
		      update_zn_flags(n);
		    }
		}


	      else if (opcode == 0x00C00000) /* SBC */
		{
		  int n = gpregs[rn] - op2;
		  if (cf==0) n -= 1;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) update_zn_flags(n);
		}
	      
	      
	      else if (opcode == 0x01400000) /* CMP */
		{
		  int op1 = gpregs[rn];
		  int n = op1 - op2;
		  if (1)
		    {
		      if (n < 0) cf = 0; else cf = 1;
		      vf = 0;
		      update_zn_flags(n);
		    }
		}


	      else if (opcode == 0x01800000) /* OR */
		{
		  int n = gpregs[rn] | op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) update_zn_flags(n);
		}
	      
	      else if (opcode == 0x01A00000) /* mov */
		{
		  if (rd == 15) gpregs[rd] = op2+4; /* another 4 gets added by oar advance */
		  else gpregs[rd] = op2;
		  if (i & (1<<20)) update_zn_flags(op2);
		}


	      else if (opcode == 0x01C00000) /* BIC */
		{
		  int n = gpregs[rn] & ~op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) update_zn_flags(n);
		}

	      else if (opcode == 0x01E00000) /* MVN */
		{
		  int n = ~ op2;
		  gpregs[rd] = n;	    
		  if (i & (1<<20)) update_zn_flags(n);
		}

	      
	      else 
		{
		  printf("Unsupported dp instruction %x\n", opcode);
		  running = 0;
		} 
	    }
	}

      else if ((i & (3 << 26))==0x04000000) /* ldr str */
	{
	  int loadf =  (i & 0x00100000);
	  int bytef = (i & 0x00400000);
	  /*  int plingf = (i & 0x00200000); */
	  int rn = (i >> 12) & 15;
	  int pref = (i & (1<<24));
	  int br = (i>>16) & 15; /* base reg */	  
	  int o, ea = gpregs[br];
	  


	  if ((i & 0x0E000000) == 0x0C000000) printf(" copro ");

	  if ((i & 0x02000000)==0)  /* indexed+offset form */
	    {
	      o = i & 0xFFF;
	      if ((i & 0x00800000)==0) o = -o; /* if downwards */

	      if (pref) ea += o;

	      if (loadf) 
		    {
		      if (rn == 15) gpregs[15] = memory_read(ea, bytef)+4;
		      else gpregs[rn] = memory_read(ea, bytef);

		    }
	      else memory_write(ea, gpregs[rn], bytef);
	    
	    }
	  else /* double reg form of load/store */
	    {
	      unsigned int shift = (i >> 4) & 0xff;
	      int o = gpregs[(i&15)];

	      if (shift) o = apply_shift(o, shift);
	      o = (i & 0x00800000) ? o:  - o;

	      if (pref) ea += o;
	      
	      if (loadf)
		{
		  /* if byte, top 24 bits are cleared */
		  if (rn == 15) gpregs[15] = memory_read(ea, bytef)+4;
		  else gpregs[rn] = memory_read(ea, bytef);
		  
		}
	      else memory_write(ea, gpregs[rn], bytef);
	    }
	    
	  if (nomemreportf == 0)
	    {
	      find_a_symbol(ea, 0);
	      putchar('\t');

	      if (loadf) printf("Loaded %s %X from EA=%08X\n", (bytef) ? "byte":"",  gpregs[rn], ea);
	      else printf(" M[%X] := %X   %c          pc=%08X\n", ea, gpregs[rn], (bytef) ? 'b': 'w', gpregs[15]-8);     
	    }

	  if (loadf) arm7_cycles += 2; else arm7_cycles += 1;

	  if (i & 0x00200000) /* write back */
	    {
	      if (!pref) ea += o;
	      gpregs[br] = ea;  /* write back */
	      
	    }
	} 
    

     else if ((i & (7<<25)) == 0x0A000000) /* branches */
       {
	 int dest;
	 int offset = i & 0xFFFFFF;
	 arm7_cycles += 2;
	 if (offset & 0x800000) offset |= 0xFF000000;
	 if (i & (1<<24)) gpregs[14] = gpregs[15]-4;  /* link */
	 dest = gpregs[15]+8 + (offset << 2);
	 gpregs[15] = dest - 4; /* gets incd below */
       }

     else if ((i & (7<<25)) == 0x08000000) /* LDM STM */
       {
	 int r, rn = (i >> 16) & 15;
	 int m = gpregs[rn];
	 int pre, post;
	 int count = 0;
	 int up;
	 for (r=0;r<16;r++) if (i & (1<<r)) count ++;
	 up = (i & (1<<23));

	 arm7_cycles += count + 1;

	 if (up==0) 
	   {
	     m -= count *4;
	     if (i & (1<<24)) { pre = 0; post = 4; } else { pre = 4; post = 0; }
	   }
	 else
	   {
	     if (i & (1<<24)) { pre = 4; post = 0; } else { pre = 0; post = 4; }
	   }
	 if (i & (1<<20)) /* if load */
	   {
	     int r;
	     for (r=0; r<16;r++)
	       {
		 if (i & (1<<r))
		   {
		     m += pre;
		     gpregs[r] = memory_read(m, 0);
		     /* If dest is PC, need to add 4 here along with normal 4 at end to make
			it 8 ahead */
		     if (r == 15)
		       {
			 arm7_cycles += 2;
			 gpregs[r] += 4;
		       }
		     m += post;
		   }
	       }
	   }
	 else /* store */
	   {
	     int r;
	     for (r=0; r<16;r++)
	       {
		 if (i & (1<<r))
		   {
		     m += pre;
		     memory_write(m, gpregs[r], 0);
		     m += post;
		   }
	       }
	   }

	 /* write back to src register */
	 if (i & (1<<21)) gpregs[rn] += (up) ? 4*count: -4*count; 


       }


      else
	{
	  printf("Unsupported instruction\n");
	  myadis(i, gpregs[15]-8);
	  running = 0;
	}
    }
  
  gpregs[15] += 4;
  arm7_total_cycles += arm7_cycles;
}

static  FILE *fd;
int iread(len)
{
  int r = 0;
  int i;
  for (i=0; i < len; i++)
  {
    unsigned char c = getc(fd);
    c = (c <= '9') ? c-'0': c-('0'+7);
    r = (r<<4) + c;
  }
  return r;
}


int read_ihex(char *fn)
{
  int line = 1;
  int addr = -1;
  int bytes = 0;  
  int segment = 0;
  int verbose = 0;
  fd = fopen(fn, "r");

  if (verbose) printf("Reading %s\n", fn);
  if (fd == 0) 
    {
      perror("cant open input file");
      exit(1);
    }

  while (feof(fd)==0)
  {
    uchar c = getc(fd);
    int llen, laddr, ltype;
    uchar lsum;
    int wq;

    if (c == '\n') line++;
    if (c != ':') continue;
    llen = iread(2);
    laddr = iread(4);
    ltype = iread(2);

    if (verbose)  printf("Type %x  addr %x len %x \n", ltype, laddr, llen);


    if (ltype == 2 && llen == 2)
      {
	uchar d;
	segment = iread(4) << 4;
	if (verbose)  printf("Segment set to %x\n", segment);
	d = iread(2);  /* dont check csum */
	continue;
      }
    else if (ltype != 0) 
    { 
      printf("Type %i encountered in line %i\n", ltype, line);
      continue;
    }

    addr = laddr + segment;
    
    lsum = llen + laddr + ltype + (laddr >> 8);
    for (wq=0; wq<llen; wq++)
    {
      uchar data = iread(2);
      lsum = lsum + data;
      if (verbose) printf("%08X %02X\n", addr+wq, data);

      bytes += 1;

      memory_write(addr+wq, data, 1);
    }
    lsum = lsum + iread(2);

    if (lsum != 0) 
          fprintf(stderr, "Checksum error in line %i, residue %i\n", line, lsum);    

    addr = addr + llen;
  }
  if (verbosef) printf("File %s ihex. Bytes read = %i bytes.\n", fn, bytes);
   
  return bytes;
}


int cbgatoi(char *s)
{
  int r = 0;
  int base = 10;
  if (s == NULL) return 0;
  while (*s == ' ' || *s == 0) s++;  /* Skip leading whities */
  while (1)
   { char c = *(s++);
     if (c >= '0' && c <= '9')
      { r = r*base + c - '0';
        continue;
      }
     if (c >= 'a') c = c - 32;
     if (c =='X')
      { base = 16;
        continue;
      }
     if (c >= 'A' && c <= 'F')
      { r = r*base + c - 55;
        continue;
      }
     break;
   }
  return r;
}


void find_a_symbol(int v, int lf)
{
  char *r;
  char line[132];
  if (symbols)
    {
      sprintf(line, "%08X", v);
      r = (char *) strstr(symbols, line);
      if (r) 
	{
	  r += 8;
	  while (*r && *r != '\n') putchar (*r++);
	  if (lf) printf("\n");
	}
    }
}



void read_symbols(char *f)
{
  int v = 100000;
  int fd = open(f, O_RDONLY);
  if (fd < 0)
    {
      printf ("could not open %s for symbols\n", f);
      return;
      symbols = NULL;
    }
  symbols = (char *) malloc(v);
  v = read(fd, symbols, v);
  symbols[v] = (char) 0;
}

void givestats()
{
  printf("Instructions %i, arm7_cycles %i, memory_reads %i, memory_writes %i\n", steps, arm7_total_cycles, mem_reads, mem_writes);
}

void read_inputfile(char *f)
{
  int v = 100000;
  int fd = open(f, O_RDONLY);
  if (fd < 0)
    {
      printf ("could not open %s for simulated uart input\n", f);
      return;
      inputfile = NULL;
    }
  inputfile = (char *) malloc(v);
  v = read(fd, inputfile, v);
  inputfile[v] = (char) 0;
}


int main(int argc, char **argv)
{
  int i;

  char *progname = argv[0];

  while (argc > 1)
    {

      if (strcmp("-byteswap", argv[1])==0)
	{
	  argc --;
	  argv ++;
	  byteswapf = 1 - byteswapf;
	  continue;
	}

      if (strcmp("-disas", argv[1])==0)
	{
	  argc --;
	  argv ++;
	  disasf = 1;
	  continue;
	}

      if (strcmp("-notihex", argv[1])==0)
	{
	  argc --;
	  argv ++;
	  ihexf = 0;
	  continue;
	}

      if (strcmp("-bp", argv[1])==0)
	{
	  breakpoint = cbgatoi(argv[2]);
	  argc -= 2;
	  argv += 2;
	  continue;
	}

      if (strcmp("-entry", argv[1])==0)
	{
	  entry_point = cbgatoi(argv[2]);
	  argc -= 2;
	  argv += 2;
	  continue;
	}

      if (strcmp("-nomemreport", argv[1])==0)
	{
	  nomemreportf = 1;
	  argc -= 1;
	  argv += 1;
	  continue;
	}

      if (strcmp("-bpt", argv[1])==0)
	{
	  breakpoint = cbgatoi(argv[2]);
	  argc -= 2;
	  argv += 2;
	  bpcont = bptype_bpt;
	  continue;
	}

      if (strcmp("-bpl", argv[1])==0)
	{
	  breakpoint = cbgatoi(argv[2]);
	  argc -= 2;
	  argv += 2;
	  bpcont = bptype_bpl;  
	  continue;
	}

      if (strcmp("-listeachstep", argv[1])==0)
	{
	  listeachstepf = 1;
	  argc -= 1;
	  argv += 1;
	  continue;
	}

      if (strcmp("-verbose", argv[1])==0)
	{
	  argc --;
	  argv ++;
	  verbosef = 1;
	  continue;
	}

      if (strcmp("-test", argv[1])==0)
	{
	  argc --;
	  argv ++;
	  myadis(0xE91da000, 0);
	  printf("\n");
	  return 0;
	}

      break;
    }

  if (argc == 1)
    {
      printf("Usenix: %s [ {-bp,-bpl,-bpt} n ] [ -entry n ]   -listeachstep -byteswap -disas -notihex -nomemreport ] fn\n", progname);
      exit(1);
    }

  read_symbols("symbols");
  read_inputfile("inputfile");
  logfile = fopen("coredump", "w");

  if (argc == 2)
    {
      int len;
      if (ihexf)
	{
	  len = read_ihex(argv[1]);
	}
      else
	{ /* binary read */
	  int f = open(argv[1], O_RDONLY);
	  if (f < 0)
	    {
	      perror("");
	      exit(1);
	    }
	  len = read(f, memory, MEMSIZE);
	  if (verbosef) printf("File %s raw binary.\nBytes read = %i bytes.\n", argv[1], len);
	}
      
      if (len == MEMSIZE) 
	{
	  printf("Input file truncated to %i bytes\n", len);
	}  
      
      
      

      if (disasf)
	{
	  for (i=LOW_ADDRESS;i<HIGH_ADDRESS;i += 4) 
	    {
	      find_a_symbol(i, 1);
	      myadis(memory_read(i, 0), i);
	     
	     }
	  printf("Disassembly complete\n");
	  return 0;
	}

      len = len >> 2;      

      gpregs[13] = START_STACK;
      gpregs[14] = 0;
      gpregs[15] = entry_point+8; /* R15 is always 8 more than it `should' be */  
      steps  = 0;
      
      while(running)
	{
	  int vb = listeachstepf;
	  step(vb); 
	  if (vb) showregs();
	  
	  if (steps > 0 && gpregs[15]==8) 
	    {
	      printf("Branch to zero termination.\n");
	      break;  /* Exit on jump back to zero */
	    }

	  if (steps++ > instruction_limit) 
	    {
	      printf("Instruction limit hit\n");
	      break;
	    }

	  if (gpregs[15]-8 == breakpoint)
	    {
	      printf("Breakpoint struck\n");
	      if (bpcont == bptype_bpl) listeachstepf = 1;
	      if (bpcont == bptype_bpt) 
		{
		  givestats();
		  showregs();
		}
	      else break;
	    }
	  
	}
      givestats();
    }

  showregs();
  printf("Stopped emulation\n");
  coredump();
  return 0;
}






/* end of cbg carmage.c */
