
/* cbg genihex .c   TTRX MAR98 */

/*
 * This file: genihex.c
 *
 *
 * Reads and splits intel hex files, etc..
 *
 * Software from Tenison Technology
 * (C) 1995 Tenison Technology 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#define uchar unsigned char

int baseaddr = 0;

const sizer = 1234567;
FILE *fd;
int oddevf = 0;
int pokef = 0;
char *opfname = NULL;

int iread(len)
{
  int r = 0;
  int i;
  for (i=0; i < len; i++)
  {
    unsigned char c = getc(fd);
    c = (c <= '9') ? c-'0': c-('0'+7);
    r = (r<<4) + c;
  }
  return r;
}


static char *pp, *base, *limit;

read_in_data(uchar c)
{
  *pp ++ = c;
}


verify(uchar c)
{
  uchar d = *pp;
  if (d != c)
    {
      printf("Verify error at 0x%x   %02x  %02x\n", pp-base, c, d);
    }
  pp ++;
  if (pp > limit) 
    {
      printf("Second file is longer");
      exit(1);
    }
}




int cbgatoi(char *s)
{
  int r = 0;
  int base = 10;
  if (s == NULL) return 0;
  while (*s == ' ' || *s == 0) s++;  /* Skip leading whities */
  while (1)
   { char c = *(s++);
     if (c >= '0' && c <= '9')
      { r = r*base + c - '0';
        continue;
      }
     if (c >= 'a') c = c - 32;
     if (c =='X')
      { base = 16;
        continue;
      }
     if (c >= 'A' && c <= 'F')
      { r = r*base + c - 55;
        continue;
      }
     break;
   }
  return r;
}




static int segment = 0;

void ihdump(unsigned char *buffer, int sa, int len)
{
  FILE *objcis = stdout;
  int i = 0;
  int la = sa;
  int amount = len;
  
  if (amount == 0) return;
  
  while (amount) 
    {
      int sum;

      int j = (amount >= 16) ? 16: amount;
      if (j + (la & 0xFFFF) > 0xFFFF) j = 0x10000 -  (la & 0xFFFF);
      amount -= j;

      if ((la & 0x000F0000) != segment)
	{
	  segment = la & 0x000F0000;
	  sum = (segment >> 12) + 2 + 2;
 	  fprintf(objcis, ":02000002%04X%02X\n", segment >> 4, (256 - sum) & 0xFF);
	  segment = la & 0x000F0000;
	}
      

      fprintf(objcis, ":%02X%04X00", j, la & 0xFFFF);
      sum = j + la + (la >> 8);
      

      la += j;
      while(j-- > 0)
	{
	  fprintf(objcis, "%02X", buffer[i]);
	  sum = buffer[i] + sum;
	  i += 1;
	}
      fprintf(objcis, "%02X\n", (256 - sum) & 255);
    }
  fprintf(objcis, "%c%c", 10, 13);
}


int main (argc, argv) char ** argv;
{
  int f;
  char * buffer = (char *) malloc(sizer);
  if (argc == 1)
    {
      printf("usenix:\n %s [-base n] [-pokef -odd -even] [-o fn]  fn\n", argv[0]);
      exit(1);
    }
  while(1)
    {

      if (!strcmp(argv[1], "-base"))
	{
	  
	  baseaddr = cbgatoi(argv[2]);
	  argc -= 2; argv += 2;
	  continue;
	}

      if (!strcmp(argv[1], "-odd"))
	{
	  oddevf = 1;
	  argc --; argv ++;
	  continue;
	}
     
      if (!strcmp(argv[1], "-poke"))
	{
	  pokef = 1;
	  argc --; argv ++;
	  continue;
	}
      if (!strcmp(argv[1], "-even"))
	{
	  oddevf = 2;
	  argc --; argv ++;
	  continue;
	}
      if (!strcmp(argv[1], "-o"))
	{
	  opfname = strdup(argv[2]);
	  argc -=2; argv +=2;
	  continue;
	}
      break;
    }
  base = buffer;
  limit = buffer+sizer;
  f = open(argv[1], O_RDONLY);
  if (f >= 0)
    {
      pp = buffer + read(f, base, sizer);
    }
  else
    {
      printf("Cant open %s\n", argv[1]);
      exit(1);
    }


  ihdump(buffer, baseaddr, pp-buffer);

  return 0;

}


int bigsum(unsigned char *l, int b)
{
  int r = 0;
  while (b-- > 0)
    {
      r += *l++;
    }
  return r;
}


run(fn, f)
char *fn;
void (*f)();
{
  int line = 1;
  int addr = -1;
  int bytes = 0;  
  fd = fopen(fn, "r");
  if (fd == 0) 
    {
      perror("cant open input file");
      exit(1);
    }

  while (feof(fd)==0)
  {
    uchar c = getc(fd);
    int llen, laddr, ltype;
    uchar lsum;
    int wq;

    if (c == '\n') line++;
    if (c != ':') continue;
    llen = iread(2);
    laddr = iread(4);
    ltype = iread(2);
/*
    printf("Type %x  addr %x len %x \n", ltype, laddr, llen);
*/
    if (ltype != 0) 
    { 
      printf("Type %i encountered in line %i\n", ltype, line);
      continue;
    }
    if (addr == -1) addr = laddr;
    else if (laddr != addr)
    {
      printf("Disjoint addresses %x and %x in line %i\n",
                   laddr, addr, line);
      addr = laddr;
    }
    lsum = llen + laddr + ltype + (laddr >> 8);
    for (wq=0; wq<llen; wq++)
    {
      uchar data = iread(2);
      lsum = lsum + data;
      f(data);
      bytes += 1;
    }
    lsum = lsum + iread(2);

    if (lsum != 0) 
          fprintf(stderr, "Checksum error in line %i, residue %i\n", line, lsum);    

    addr = addr + llen;
  }
  /*  printf("Bytes read = %i bytes.\n", bytes);
   */
  return 0;
}

/* eof */
